#ifndef _twavefileout_
#define _twavefileout_

#include "TWaveIO.hpp"

#define wRIFF 0
#define wWAVE 1
#define wFMT  2
#define wC16  3
#define wFACT 4
#define wDATA 5

static const char *strs[] = {
   "RIFF",              // 0
   "WAVE",              // 1
   "fmt ",              // 2
   "                ",  // 3
   "fact",              // 4
   "data"               // 5
};

class wavHeader {
    const char          *ckID;
    ui32                ckSize;
    const char          *WAVEID;
public:
    wavHeader() {
        ckID   = strs[wRIFF];
        WAVEID = strs[wWAVE];
    }
    void setSize( ui32 s ) {
        ckSize = s;
    }
    ui32 Write( FILE *fp ) {
        return -8+
            fwrite( ckID,    1, 4, fp ) +
            fwrite( &ckSize, 1, 4, fp ) +
            fwrite( WAVEID,  1, 4, fp );
    }
};

class wavFormat {
public:
    const char          *ckID;
    ui32                ckSize;
    ui16                wFormatTag;
    ui16                nChannels;
    ui32                nSamplesPerSec;
    ui32                nAvgBytesPerSec;
    ui32                nBlockAlign;
    ui16                wBitsPerSample;
    ui16                cbSize;
    ui16                wValidBitsPerSample;
    ui32                dwChannelMask;
    const char          *SubFormat;

    wavFormat() {
        ckID      = strs[wFMT];
        ckSize    = 16;
        wFormatTag = WAVE_FORMAT_PCM;
        SubFormat = strs[wC16];
    }
    ui32 Write( FILE *fp ) {
        return
            fwrite( ckID,                       1, 4, fp ) +
            fwrite( &ckSize,                    1, 4, fp ) +
            fwrite( &wFormatTag,                1, 2, fp ) +
            fwrite( &nChannels,                 1, 2, fp ) +
            fwrite( &nSamplesPerSec,            1, 4, fp ) +
            fwrite( &nAvgBytesPerSec,           1, 4, fp ) +
            fwrite( &nBlockAlign,               1, 2, fp ) +
            fwrite( &wBitsPerSample,            1, 2, fp ); /* +
            fwrite( &cbSize,                    1, 2, fp ) +
            fwrite( &wValidBitsPerSample,       1, 2, fp ) +
            fwrite( &dwChannelMask,             1, 4, fp ) +
            fwrite( &SubFormat,                 1, 16, fp );  */
    }
};

class wavFact {
    const char          *ckID;
    ui32                ckSize;
public:
    ui32                dwSampleLength;
    wavFact() {
        ckID      = strs[wFACT];
        ckSize    = 4;
    }
    ui32 Write( FILE *fp ) {
        return
            fwrite( ckID,                       1, 4, fp ) +
            fwrite( &ckSize,                    1, 4, fp ) +
            fwrite( &dwSampleLength,            1, 4, fp );
    }
};

class wavData {
    const char          *ckID;
public:
    wavData() {
        ckID      = strs[wDATA];
    }
    ui32 Write( FILE *fp, ui32 len=0 ) {
        return
            fwrite( ckID,                       1, 4, fp ) +
            fwrite( &len,                       1, 4, fp );

    }
};

class TWaveFileOut {
    wavHeader   wh;
    wavFormat   wf;
    FILE        *fp;
    ui32        BytesWritten;
    ui32        DataStart;
    ui32        sampdata;
public:
    wavData     wd;
    
    FILE *Open(
        char        *Filename,
        ui16        wFormatTag,
        ui16        nChannels,
        ui32        nSamplesPerSec,
        ui16        wBitsPerSample
    ) {
        if( !fp ) {
            wf.wFormatTag       = wFormatTag;
            wf.nChannels        = nChannels;
            wf.nSamplesPerSec   = nSamplesPerSec;
            wf.wBitsPerSample   = wBitsPerSample;

            wf.nBlockAlign      = wf.nChannels * wf.wBitsPerSample / 8;
            wf.nAvgBytesPerSec  = wf.nSamplesPerSec * wf.nBlockAlign;

            fp = fopen( Filename, "wb" );
            if( fp ) {
                BytesWritten  = wh.Write( fp );
                BytesWritten += wf.Write( fp );
                DataStart     = 8+BytesWritten;
                BytesWritten += wd.Write( fp );

                sampdata  = 0;
            }
        }
        return fp;
    }
    void Close() {
        if( fp ) {
            fwrite( "", 1, (sampdata&1), fp );

            fseek( fp, 0, SEEK_SET );
            wh.setSize( BytesWritten );
            wh.Write( fp );

            fseek( fp, DataStart, SEEK_SET );
            wd.Write( fp, sampdata );
            fclose( fp );
            fp = NULL;
        }
    }
    TWaveFileOut() {
        fp = NULL;
    }
    TWaveFileOut(
        char        *Filename,
        ui16        wFormatTag,
        ui16        nChannels,
        ui32        nSamplesPerSec,
        ui16        wBitsPerSample
    ) {
        fp = NULL;
        Open(
            Filename,
            wFormatTag,
            nChannels,
            nSamplesPerSec,
            wBitsPerSample
        );
    }
    ~TWaveFileOut() {
        Close();
    }
    void Write( unsigned char *data, ui32 len ) {
        sampdata += len;
        BytesWritten += fwrite( data, 1, len, fp );
    }
    inline int Flush() {
        return fflush( fp );
    }
};

#endif