#ifndef _twavein_
#define _twavein_

#include "TWaveIO.hpp"

class TWaveIn {
private:
    unsigned int        uHeaders;
    WAVEFORMATEX        pwfx; // Could be used and thrown away but saved for querying
    HANDLE              BufDoneEvent;

    friend void CALLBACK TWaveInProc( HWAVEIN hwo,UINT uMsg,DWORD dwInstance,
                                      DWORD dwParam1,DWORD dwParam2 );
public:

    HWAVEOUT hwo;
    WAVEHDR  *wh;
    DWORD    dwBufferLength;
    UINT     uDeviceID;
    unsigned int iHeadR;
    char     *CurBufPtr;

    TWaveIn( unsigned int Headers=2 ) {
        if( Headers<2 ) Headers = 2;

        uHeaders = Headers;
        wh = new WAVEHDR[uHeaders];


        CurBufPtr = NULL;

        BufDoneEvent = CreateEvent(NULL,FALSE,false,NULL);

        pwfx.nSamplesPerSec = 44100;
    }
    ~TWaveIn() {
        delete []wh;
        CloseHandle( BufDoneEvent );
    }
    inline UINT GetNumDevs() {
        return waveInGetNumDevs();
    }
    GetDevCaps( UINT did, WAVEINCAPS *wic ) {
        return waveInGetDevCaps( did, wic, sizeof( WAVEINCAPS ) );
    }
    inline DWORD GetSamplesPerSecond() {
        return pwfx.nSamplesPerSec;
    }
    MMRESULT OpenPCM( UINT uDeviceID=WAVE_MAPPER, DWORD SamplesPerBuf=8192, DWORD sps=44100,
                      WORD chan=2, WORD bps=16 )
    {
        MMRESULT retval;

        DWORD dwCallback         = (DWORD) TWaveInProc;
        DWORD dwCallbackInstance = 0;
        DWORD fdwOpen            = CALLBACK_FUNCTION ; //| WAVE_ALLOWSYNC;


        pwfx.wFormatTag      = WAVE_FORMAT_PCM;
        pwfx.nChannels       = chan;
        pwfx.nSamplesPerSec  = sps;
        pwfx.wBitsPerSample  = bps;
        pwfx.cbSize          = 0;
        pwfx.nBlockAlign     = (WORD) (pwfx.nChannels*pwfx.wBitsPerSample/8); // Bytes per sample
        pwfx.nAvgBytesPerSec = pwfx.nSamplesPerSec*pwfx.nBlockAlign;

        dwBufferLength = SamplesPerBuf*pwfx.nBlockAlign;

        retval = waveInOpen(
            &hwo,
            uDeviceID,
            &pwfx,
            dwCallback,
            (DWORD) this,
            fdwOpen );

        if( retval == MMSYSERR_NOERROR ) {
            unsigned int hdr;

            for( hdr=0; hdr<uHeaders && retval==MMSYSERR_NOERROR; hdr++ ) {
                wh[hdr].lpData = new char[dwBufferLength];
                wh[hdr].dwBufferLength = dwBufferLength;
                wh[hdr].dwBytesRecorded = 0;
                wh[hdr].dwUser = 0;
                wh[hdr].dwFlags = 0; // WHDR_BEGINLOOP & WHDR_ENDLOOP; // 0
                wh[hdr].dwLoops = 0;

                retval = waveInPrepareHeader( hwo, &wh[hdr], sizeof( WAVEHDR ) );
                if( retval == MMSYSERR_NOERROR ) {
                    retval = waveInAddBuffer(hwo,&wh[hdr], sizeof( WAVEHDR ) );
                }
            }
            iHeadR = 0; // Start reading here

            CurBufPtr = wh[iHeadR].lpData;
            waveInStart(hwo);
        }
        return retval;
    }
    MMRESULT Close() {
        unsigned int hdr;
        MMRESULT retval=MMSYSERR_NOERROR;


        waveInReset( hwo );
        WaitUntilAllDone();

        for( hdr=0; hdr<uHeaders && retval==MMSYSERR_NOERROR; hdr++ ) {
            retval = waveInUnprepareHeader( hwo, &wh[hdr], sizeof( WAVEHDR ) );
            if( retval==MMSYSERR_NOERROR ) {
                delete []wh[hdr].lpData;
            }
        }
        if(retval==MMSYSERR_NOERROR) retval=waveInClose( hwo );
        return retval;
    }
    DWORD ReadBuffer( char *&buffer, DWORD &length, DWORD dwMilliseconds=INFINITE ) {
        DWORD retval = WAIT_OBJECT_0;

        if( !(wh[iHeadR].dwFlags&WHDR_DONE) ) retval = WaitForSingleObject( BufDoneEvent, dwMilliseconds );

        buffer = wh[iHeadR].lpData;
        length = wh[iHeadR].dwBytesRecorded / pwfx.nBlockAlign; // samples
        waveInAddBuffer(hwo, &wh[iHeadR], sizeof( WAVEHDR ) );
        iHeadR=(iHeadR+1)%uHeaders;

        return retval; // should be == WAIT_OBJECT_0
    }
    i16 Read_16() {   // either use ONLY this OR ReadBuffer;
        static i16 retval;
        if( !(wh[iHeadR].dwFlags&WHDR_DONE) ) {
            WaitForSingleObject( BufDoneEvent, INFINITE );
        }
        retval = *((i16*)CurBufPtr);
        CurBufPtr += sizeof( i16 );
        if( (DWORD)(CurBufPtr-wh[iHeadR].lpData)>=wh[iHeadR].dwBytesRecorded ) {
            waveInAddBuffer(hwo, &wh[iHeadR], sizeof( WAVEHDR ) );
            iHeadR=(iHeadR+1)%uHeaders;
            CurBufPtr = wh[iHeadR].lpData;
        }
        return retval;
    }






    void WaitUntilAllDone() { // wait for all the buffers to be returned
        while( !(wh[iHeadR].dwFlags&WHDR_DONE) ) {
            WaitForSingleObject( BufDoneEvent, INFINITE );
        }
    }
};










void CALLBACK TWaveInProc(
        HWAVEIN hwo,
        UINT uMsg,
        DWORD dwInstance,
        DWORD dwParam1,
        DWORD dwParam2 )
{
    static TWaveIn *cl;
    if( uMsg==WIM_DATA ) {
        cl = (TWaveIn*) dwInstance;
        SetEvent( cl->BufDoneEvent );
    }
}

#endif