#ifndef _twaveout_
#define _twaveout_

#include "TWaveIO.hpp"

class TWaveOut {
private:
    unsigned int        uHeaders;
    WAVEFORMATEX        pwfx; // Could be used and thrown away but saved for querying
    HANDLE              BufDoneEvent;
    bool                FillAllBuffs;
    friend void CALLBACK TWaveOutProc( HWAVEOUT hwo,UINT uMsg,DWORD dwInstance,
                                       DWORD dwParam1,DWORD dwParam2 );
public:
    bool     Paused;
    HWAVEOUT hwo;
    WAVEHDR  *wh;
    DWORD    dwBufferLength;
    UINT     uDeviceID;
    unsigned int iHeadW, iHeadR;
    char     *CurBufPtr;

    TWaveOut( unsigned int Headers=2 ) {
        if( Headers<2 ) Headers = 2;

        uHeaders = Headers;
        wh = new WAVEHDR[uHeaders];

        Paused = true;
        CurBufPtr = NULL;

        BufDoneEvent = CreateEvent(NULL,FALSE,false,NULL);

        pwfx.nSamplesPerSec = 44100;
    }
    ~TWaveOut() {
        delete []wh;
        CloseHandle( BufDoneEvent );
    }
    inline UINT GetNumDevs() {
        return waveOutGetNumDevs();
    }
    GetDevCaps( UINT did, WAVEOUTCAPS *woc ) {
        return waveOutGetDevCaps( did, woc, sizeof( WAVEOUTCAPS ) );
    }
    inline DWORD GetSamplesPerSecond() {
        return pwfx.nSamplesPerSec;
    }
    MMRESULT OpenPCM( UINT uDeviceID=WAVE_MAPPER, DWORD SamplesPerBuf=8192, DWORD sps=44100,
                      WORD chan=2, WORD bps=16 )
    {
        MMRESULT retval;

        DWORD dwCallback         = (DWORD) TWaveOutProc;
        DWORD dwCallbackInstance = 0;
        DWORD fdwOpen            = CALLBACK_FUNCTION; // | WAVE_ALLOWSYNC;
        FillAllBuffs             = true;

        pwfx.wFormatTag      = WAVE_FORMAT_PCM;
        pwfx.nChannels       = chan;
        pwfx.nSamplesPerSec  = sps;
        pwfx.wBitsPerSample  = bps;
        pwfx.cbSize          = 0;
        pwfx.nBlockAlign     = (WORD) (pwfx.nChannels*pwfx.wBitsPerSample/8); // Bytes per sample
        pwfx.nAvgBytesPerSec = pwfx.nSamplesPerSec*pwfx.nBlockAlign;

        dwBufferLength = SamplesPerBuf*pwfx.nBlockAlign;

        retval = waveOutOpen(
            &hwo,
            uDeviceID,
            &pwfx,
            dwCallback,
            (DWORD) this,
            fdwOpen );

        if( retval == MMSYSERR_NOERROR ) {
            unsigned int hdr;

            for( hdr=0; hdr<uHeaders && retval==MMSYSERR_NOERROR; hdr++ ) {
                wh[hdr].lpData = new char[dwBufferLength];
                wh[hdr].dwBufferLength = dwBufferLength;
                wh[hdr].dwBytesRecorded = 0;
                wh[hdr].dwUser = 0;
                wh[hdr].dwFlags = 0; // WHDR_BEGINLOOP & WHDR_ENDLOOP; // 0
                wh[hdr].dwLoops = 1;

                retval = waveOutPrepareHeader( hwo, &wh[hdr], sizeof( WAVEHDR ) );



            }
            iHeadR = 0;
            iHeadW = 0; // Start writing here
            CurBufPtr = wh[iHeadW].lpData;

        }
        return retval;
    }
    MMRESULT Close() {
        unsigned int hdr;
        MMRESULT retval=MMSYSERR_NOERROR;

        Flush();
        WaitUntilAllDone();
        waveOutReset( hwo ); // maybe not needed, all buffers done

        for( hdr=0; hdr<uHeaders && retval==MMSYSERR_NOERROR; hdr++ ) {
            retval = waveOutUnprepareHeader( hwo, &wh[hdr], sizeof( WAVEHDR ) );
            if( retval==MMSYSERR_NOERROR ) {
                delete []wh[hdr].lpData;
            }
        }
        if(retval==MMSYSERR_NOERROR) retval=waveOutClose( hwo );
        return retval;
    }
    DWORD WriteBuffer( char *buffer, DWORD length, DWORD dwMilliseconds=INFINITE ) {  // NOT FINSIHED!!!
        DWORD retval = WAIT_OBJECT_0;
        
        if( (wh[iHeadW].dwFlags&WHDR_INQUEUE) ) WaitForSingleObject( BufDoneEvent, dwMilliseconds );

        buffer = wh[iHeadR].lpData;
        length = wh[iHeadR].dwBytesRecorded / pwfx.nBlockAlign; // samples
        waveInAddBuffer(hwo, &wh[iHeadR], sizeof( WAVEHDR ) );
        iHeadR=(iHeadR+1)%uHeaders;

        return retval; // should be == WAIT_OBJECT_0
    }
    void Write_16( i16 Sample ) {
        if( (wh[iHeadW].dwFlags&WHDR_INQUEUE) ) {
            WaitForSingleObject( BufDoneEvent, INFINITE );
        }
        *((i16*)CurBufPtr) = Sample;
        CurBufPtr += sizeof( i16 );
        if( (DWORD)(CurBufPtr-wh[iHeadW].lpData)>=dwBufferLength ) { // buffer full
            /*
            if( FillAllBuffs && ((iHeadW+1)%uHeaders)==iHeadR ) { // first round, fill all buffers before sending
                FillAllBuffs=false;
                do {
                    waveOutWrite( hwo, &wh[iHeadR], sizeof( WAVEHDR ) );
                    iHeadR=(iHeadR+1)%uHeaders;
                } while( !(wh[iHeadR].dwFlags&WHDR_INQUEUE) );
            } else {
            */
                waveOutWrite( hwo, &wh[iHeadW], sizeof( WAVEHDR ) );
            // }
            iHeadW=(iHeadW+1)%uHeaders;
            CurBufPtr = wh[iHeadW].lpData;
        }
    }
    void WaitUntilAllDone() { // wait for all the buffers to be returned
        while( (wh[iHeadR].dwFlags&WHDR_INQUEUE) ) {
            WaitForSingleObject( BufDoneEvent, INFINITE );
        }
    }
    void Flush() { // send any unsent halfwritten buffers.
        if( !(wh[iHeadW].dwFlags&WHDR_INQUEUE) && CurBufPtr != wh[iHeadW].lpData ) {
            wh[iHeadW].dwBufferLength = CurBufPtr - wh[iHeadW].lpData;
            waveOutWrite( hwo, &wh[iHeadW], sizeof( WAVEHDR ) );
            iHeadW=(iHeadW+1)%uHeaders;
            CurBufPtr = wh[iHeadW].lpData;
            FillAllBuffs = true;
        }
    }
};

void CALLBACK TWaveOutProc(
        HWAVEOUT hwo,
        UINT uMsg,
        DWORD dwInstance,
        DWORD dwParam1,
        DWORD dwParam2 )
{
    static TWaveOut *cl;
    if( uMsg==WOM_DONE ) {
        cl = (TWaveOut*) dwInstance;

        cl->wh[cl->iHeadR].dwBufferLength = cl->dwBufferLength;
        cl->iHeadR = (cl->iHeadR+1)%cl->uHeaders;
        cl->Flush();
        SetEvent( cl->BufDoneEvent );
    }
}

#endif